# coding: utf-8
from __future__ import print_function, division, absolute_import, unicode_literals

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from .helpers import get_object_repr


ADDITION = 1
CHANGE = 2
DELETION = 3

ACTION_TYPE_CHOICES = ((ADDITION, _("Addition")), (CHANGE, _("Change")), (DELETION, _("Deletion")))

ACTION_MESSAGES = {
    ADDITION: _("""{} has added '{}'"""),
    CHANGE: _("""{} has updated '{}'"""),
    DELETION: _("""{} has deleted '{}'"""),
}


@python_2_unicode_compatible
class Entry(models.Model):
    created_date = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="history_entries", null=True, blank=True)

    object_type = models.ForeignKey(ContentType, blank=True, null=True)
    object_id = models.PositiveIntegerField(blank=True, null=True)
    content_object = GenericForeignKey("object_type", "object_id")
    object_repr = models.CharField(max_length=200)

    # The short message for humans
    action_type = models.PositiveSmallIntegerField(choices=ACTION_TYPE_CHOICES)
    short_message = models.TextField(null=True, blank=True)

    class Meta:
        ordering = ("-created_date",)

    def __str__(self):
        return self.short_message

    def is_addition(self):
        return self.action_type == ADDITION

    def is_change(self):
        return self.action_type == CHANGE

    def is_deletion(self):
        return self.action_type == DELETION

    def get_user_full_name(self):
        if self.created_by:
            return self.created_by.get_full_name() or self.created_by.username
        return _("Anonymous")

    @staticmethod
    def create(user, content_object, action_type, changelog=None, object_repr=""):
        changelog = changelog or []
        entry = Entry(created_by=user, action_type=action_type)
        if content_object:
            entry.object_type = ContentType.objects.get_for_model(content_object)
            entry.object_id = content_object.id
            entry.object_repr = get_object_repr(content_object)
        else:
            entry.object_repr = object_repr
        entry.short_message = ACTION_MESSAGES[action_type].format(entry.get_user_full_name(), entry.object_repr)
        entry.save()

        [ChangedData(entry=entry, **data).save() for data in changelog]
        return entry


class ChangedData(models.Model):
    entry = models.ForeignKey(Entry)
    label = models.CharField(_("Label"), max_length=200)
    initial_value = models.TextField(_("Initial value"), blank=True, null=True)
    changed_value = models.TextField(_("Updated value"), blank=True, null=True)


class HistoryBaseModel(models.Model):
    """Base model that implements utilities to query history entries of your models.
    Just extends your model with this one :

    >> class MyModel(HistoryBaseModel):
    >>     pass
    >>
    >> inst = MyModel.objects.get(...)
    >> inst.get_history_entries()
    >> # Returns <Entry queryset>

    """

    def get_history_entries(self):
        return Entry.objects.filter(object_id=self.pk, object_type=ContentType.objects.get_for_model(self))

    class Meta:
        abstract = True



# coding: utf-8
from __future__ import print_function, division, absolute_import, unicode_literals
import os

from django.apps import apps
from django.contrib.admin.sites import AdminSite
from django.contrib.auth import get_user_model
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase, RequestFactory

from modelformhistory.admin import HistoryModelAdmin
from modelformhistory.apps import ModelformhistoryConfig

from modelformhistory.forms import HistoryModelFormMixin
from modelformhistory.models import Entry, ADDITION
from sampleapp.models import Foo, Bar, Baz
from sampleapp.forms import FooModelForm, FooModelFormRequest


User = get_user_model()

__all__ = ("ModelFormHistoryTestCase", "ModelFormHistoryAdminTestCase")


class MockRequest:
    pass


class MockSuperUser:
    def has_perm(self, perm):
        return True


request = MockRequest()
request.user = MockSuperUser()


class ModelFormHistoryAdminTestCase(TestCase):
    def setUp(self):
        bar = Bar.objects.create(name="bar")
        self.foo = Foo.objects.create(name="Test foo", integer=1, choose_somthing="ok", bar=bar)
        self.site = AdminSite()

    def test_modeladmin_str(self):
        ma = HistoryModelAdmin(Foo, self.site)
        form = ma.get_form(request)()
        self.assertIsInstance(form, HistoryModelFormMixin)


class ModelFormHistoryTestCase(TestCase):
    def setUp(self):
        self.user = User(username="test_user", last_name="TEST", first_name="User")
        self.user.save()

        bars = ["bar", "rab", "bra"]
        bazs = ["baz", "zab", "bza"]
        [Bar(name=bar).save() for bar in bars]
        [Baz(name=baz).save() for baz in bazs]
        bar = Bar.objects.get(name="bar")
        bazs = Baz.objects.all()

        self.foo = Foo(name="Test foo", integer=1, choose_somthing="ok", bar=bar)
        self.foo.save()
        [self.foo.baz.add(b) for b in bazs]
        self.indentical_datas = {
            "bar": self.foo.bar.id,
            "name": self.foo.name,
            "baz": [str(baz.id) for baz in self.foo.baz.all()],
            "integer": self.foo.integer,
            "choose_somthing": self.foo.choose_somthing,
            "yesorno": "1",
        }

    def check_changed_data(self, changed_data, label, initial_value, changed_value):
        self.assertEqual(changed_data.label, label)
        self.assertEqual(changed_data.initial_value, initial_value)
        self.assertEqual(changed_data.changed_value, changed_value)

    def test_str(self):
        entry = Entry.create(self.user, self.foo, ADDITION, changelog=[])
        self.assertEqual(str(entry), "User TEST has added 'Foo Test foo'")

    def test_no_change(self):
        form = FooModelForm(user=self.user, instance=self.foo, data=self.indentical_datas)
        form.save()
        self.assertEqual(Entry.objects.all().count(), 0)

    def test_update_choicefield(self):
        data = self.indentical_datas.copy()

        # Save another value on the choiceField
        data["choose_somthing"] = "nok"
        form = FooModelForm(user=self.user, instance=self.foo, data=data)
        form.save()
        self.assertEqual(Entry.objects.all().count(), 1)
        entry = Entry.objects.all()[0]
        self.assertTrue(entry.is_change())
        self.assertFalse(entry.is_addition())
        self.assertFalse(entry.is_deletion())
        self.assertEqual(entry.changeddata_set.all().count(), 1)
        changed_data = entry.changeddata_set.all()[0]
        self.check_changed_data(changed_data, "Make your choice", "It's OK", "It's not OK")

    def test_update_foreign_key(self):
        data = self.indentical_datas.copy()

        # Save another value on the foreignKey
        data["bar"] = Bar.objects.get(name="rab").id
        form = FooModelForm(user=self.user, instance=self.foo, data=data)
        form.save()
        self.assertEqual(Entry.objects.all().count(), 1)
        entry = Entry.objects.all()[0]
        self.assertTrue(entry.is_change())
        self.assertFalse(entry.is_addition())
        self.assertFalse(entry.is_deletion())
        self.assertEqual(entry.changeddata_set.all().count(), 1)
        changed_data = entry.changeddata_set.all()[0]
        self.check_changed_data(changed_data, "Name of the bar", "bar", "rab")

        # Save an empty value on the foreignKey
        del data["bar"]
        form = FooModelForm(user=self.user, instance=self.foo, data=data)
        form.save()
        self.assertEqual(Entry.objects.all().count(), 2)
        entry = Entry.objects.all()[0]
        self.assertEqual(entry.changeddata_set.all().count(), 1)
        changed_data = entry.changeddata_set.all()[0]
        self.check_changed_data(changed_data, "Name of the bar", "rab", "Empty")

    def test_update_manytomany(self):
        data = self.indentical_datas.copy()

        # Save another multiple values on M2M
        data["baz"] = [b.id for b in Baz.objects.filter(name__startswith="b")]
        form = FooModelForm(user=self.user, instance=self.foo, data=data)
        form.save()
        self.assertEqual(Entry.objects.all().count(), 1)
        entry = Entry.objects.all()[0]
        self.assertTrue(entry.is_change())
        self.assertFalse(entry.is_addition())
        self.assertFalse(entry.is_deletion())
        self.assertEqual(entry.changeddata_set.all().count(), 1)
        changed_data = entry.changeddata_set.all()[0]
        self.check_changed_data(changed_data, "Select some baz", "baz, zab, bza", "baz, bza")

        # Save a single object on M2M
        data["baz"] = [Baz.objects.get(name="zab").id]
        form = FooModelForm(user=self.user, instance=self.foo, data=data)
        form.save()
        self.assertEqual(Entry.objects.all().count(), 2)
        entry = Entry.objects.all()[0]
        changed_data = entry.changeddata_set.all()[0]
        self.assertEqual(entry.changeddata_set.all().count(), 1)
        self.check_changed_data(changed_data, "Select some baz", "baz, bza", "zab")

        # Save an empty value on M2M
        del data["baz"]
        form = FooModelForm(user=self.user, instance=self.foo, data=data)
        form.save()
        self.assertEqual(Entry.objects.all().count(), 3)
        entry = Entry.objects.all()[0]
        self.assertEqual(entry.changeddata_set.all().count(), 1)
        changed_data = entry.changeddata_set.all()[0]
        self.check_changed_data(changed_data, "Select some baz", "zab", "Empty")

    def test_update_boolean(self):
        data = self.indentical_datas.copy()

        data["yesorno"] = ""
        form = FooModelForm(user=self.user, instance=self.foo, data=data)
        form.save()
        self.assertEqual(Entry.objects.all().count(), 1)
        entry = Entry.objects.all()[0]
        changed_data = entry.changeddata_set.all()[0]
        self.check_changed_data(changed_data, "Check for yes", "Yes", "No")

    def test_update_filefield(self):
        data = self.indentical_datas.copy()
        upload_file = open("images/poney.jpeg", "rb")
        file_dict = {"picture": SimpleUploadedFile(upload_file.name, upload_file.read())}
        form = FooModelForm(user=self.user, instance=self.foo, data=data, files=file_dict)
        self.assertTrue(form.is_valid())
        form.save()
        self.assertEqual(Entry.objects.all().count(), 1)
        entry = Entry.objects.all()[0]
        self.assertEqual(entry.changeddata_set.all().count(), 1)
        changed_data = entry.changeddata_set.all()[0]
        self.check_changed_data(changed_data, "Picture", None, "poney.jpeg")

        upload_file = open("images/poney-bw.jpg", "rb")
        file_dict = {"picture": SimpleUploadedFile(upload_file.name, upload_file.read())}
        form = FooModelForm(user=self.user, instance=self.foo, data=data, files=file_dict)
        self.assertTrue(form.is_valid())
        form.save()
        self.assertEqual(Entry.objects.all().count(), 2)
        entry = Entry.objects.all()[0]
        self.assertEqual(entry.changeddata_set.all().count(), 1)
        changed_data = entry.changeddata_set.all()[0]
        self.check_changed_data(changed_data, "Picture", "poney.jpeg", "poney-bw.jpg")

        os.remove("poney.jpeg")
        os.remove("poney-bw.jpg")

    def test_add(self):
        data = self.indentical_datas.copy()
        form = FooModelForm(user=self.user, instance=None, data=data)
        form.save()
        self.assertEqual(Entry.objects.all().count(), 1)
        entry = Entry.objects.all()[0]
        self.assertEqual(entry.changeddata_set.all().count(), 0)
        self.assertTrue(entry.is_addition())
        self.assertFalse(entry.is_change())
        self.assertEqual(entry.short_message, "User TEST has added 'Foo Test foo'")

    def test_get_history_user(self):
        form = FooModelForm(user=self.user, instance=None, data=self.indentical_datas)
        self.assertEqual(form.get_history_user(), self.user)
        form = FooModelFormRequest(request=None, instance=None, data=self.indentical_datas)
        self.assertEqual(form.get_history_user(), None)
        request = RequestFactory()
        request.user = self.user
        form = FooModelFormRequest(request=request, instance=None, data=self.indentical_datas)
        self.assertEqual(form.get_history_user(), self.user)

    def test_get_history_entries(self):
        self.assertEqual(self.foo.get_history_entries().count(), 0)
        data = self.indentical_datas.copy()
        data["name"] = "Another name"
        form = FooModelForm(user=self.user, instance=self.foo, data=data)
        form.save()
        self.assertEqual(self.foo.get_history_entries().count(), 1)
        entry = self.foo.get_history_entries()[0]
        self.assertEqual(entry.content_object, self.foo)
        self.assertEqual(entry.changeddata_set.all().count(), 1)


class ModelformhistoryConfigTest(TestCase):
    def test_apps(self):
        self.assertEqual(ModelformhistoryConfig.name, "modelformhistory")
        self.assertEqual(apps.get_app_config("modelformhistory").name, "modelformhistory")



import os
from django.utils.translation import ugettext_lazy as _

DEBUG = True
SECRET_KEY = "tests"
ROOT_URLCONF = "tests.urls"
DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}
SESSION_ENGINE = "django.contrib.sessions.backends.cache"
INSTALLED_APPS = (
    "django.contrib.sessions",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "sampleapp",
    "modelformhistory",
    "django_nose",
)
# TEST_RUNNER = "django_nose.NoseTestSuiteRunner"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "DIRS": [os.path.dirname(__file__)],
        "OPTIONS": {
            "context_processors": (
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
            )
        },
    }
]
MIDDLEWARE = (
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
)
LOGIN_REDIRECT_URL = "/"

USE_I18N = True
LANGUAGE_CODE = "en"

LANGUAGES = (("en", _("English")), ("fr", _("French")))
